@@ -12,4 +12,11 @@ import presto.controller
 import presto.checks
 import presto.reporters
 import presto.potentials
-import presto.quasiclassical
+import presto.initialization
+import presto.external
+import presto.termination
+
+import logging
+logger = logging.getLogger()
+
+logger.info(f"presto initialized successfully - all modules loaded")
