@@ -33,6 +33,13 @@ class Trajectory():
         checkpoint_interval (int):
         lock (fasteners.InterProcessLock): lock object
         save_interval (int): how many frames to save
+        buffer (int): how many frames to keep in memory
+
+        bath_scheduler (function): maps time to desired temperature, in K.
+        termination_function (function): determines if the trajectory is finished or not
+
+        target_pressure (float): desired pressure for NPT simulation, or ``None`` for NVT simulation. in atm.
+        barostat_time_constant (float): desired time constant for NPT simulation barostat coupling. in fs.
     """
 
     def __init__(
@@ -49,7 +56,12 @@ class Trajectory():
         checkpoint_interval=10,
         stop_time=None,
         save_interval=1,
+        buffer=100,
         load_frames="all", # or ``first`` or ``last`` or a slice
+        bath_scheduler=298,
+        termination_function=None,
+        target_pressure=None,
+        barostat_time_constant=100,
         **kwargs
     ):
 
@@ -58,6 +70,13 @@ class Trajectory():
             assert timestep > 0, "can't have timestep ≤ 0!"
             self.timestep = float(timestep)
 
+        # also do this first, so checkpoint file can overrule as needed
+        if forwards is not None:
+            assert isinstance(forwards, bool), "forwards must be bool"
+            self.forwards = forwards
+        elif not hasattr(self, "forwards"):
+            self.forwards = True
+
         if checkpoint_filename is not None:
             assert isinstance(checkpoint_filename, str), "need string for file"
         self.checkpoint_filename = checkpoint_filename
@@ -65,21 +84,18 @@ class Trajectory():
         assert isinstance(checkpoint_interval, int) and checkpoint_interval > 0, "checkpoint_interval must be positive integer"
         self.checkpoint_interval = checkpoint_interval
 
+        assert isinstance(save_interval, int), "save_interval needs to be positive"
+        assert save_interval > 0, "save_interval needs to be positive"
+        self.save_interval = save_interval
+
         self.lock = None
         self.initialize_lock()
         self.frames = list()
 
         if self.has_checkpoint():
-            if load_frames=="all":
-                self.load_from_checkpoint(slice(None))
-            elif load_frames=="first":
-                self.load_from_checkpoint(slice(1,None,None))
-            elif load_frames=="last":
-                self.load_from_checkpoint(slice(-1,None,None))
-            else:
-                assert isinstance(load_frames, slice), "load_frames must be ``all``, ``first``, ``last``, or slice"
-                self.load_from_checkpoint(load_frames)
+            self.load_from_checkpoint(load_frames)
 
+        # now we carry on building the "mundane" attributes 
         if calculator is not None:
             assert isinstance(calculator, presto.calculators.Calculator), "need a valid calculator!"
         self.calculator = calculator
@@ -126,20 +142,43 @@ class Trajectory():
         if not hasattr(self, "frames"):
             self.frames = []
 
-        if forwards is not None:
-            assert isinstance(forwards, bool), "forwards must be bool"
-            self.forwards = forwards
-        elif not hasattr(self, "forwards"):
-            self.forwards = True
-
         if not hasattr(self, "stop_time"):
             assert (isinstance(stop_time, float)) or (isinstance(stop_time, int)), "stop_time needs to be numeric!"
             assert stop_time > 0, "stop_time needs to be positive!"
             self.stop_time = stop_time
 
-        assert isinstance(save_interval, int), "save_interval needs to be positive"
-        assert save_interval > 0, "save_interval needs to be positive"
-        self.save_interval = save_interval
+        assert isinstance(buffer, int), "buffer needs to be positive"
+        assert buffer > 0, "buffer needs to be positive"
+        self.buffer = buffer
+
+        # build bath scheduler
+        if hasattr(bath_scheduler, "__call__"):
+            self.bath_scheduler = bath_scheduler
+        elif isinstance(bath_scheduler, (int, float)):
+            # most of the time it's ok just to keep things constant.
+            def sched(time):
+                return bath_scheduler
+            self.bath_scheduler = sched
+        else:
+            raise ValueError(f"unknown type {type(bath_scheduler)} for bath_scheduler - want either a function or a number!")
+
+        # build termination function
+        if termination_function is not None:
+            assert hasattr(termination_function, "__call__"), "termination_function must be a function!"
+            self.termination_function = termination_function
+        else:
+            # if we haven't specified any criteria, we don't want to end before time's up! so we'll just say "end never."
+            def term(time):
+                return False
+            self.termination_function = term
+
+        # build barostat
+        if target_pressure is not None:
+            assert isinstance(target_pressure, (int, float)), "target_pressure must be ``None`` or numeric"
+            self.target_pressure = target_pressure
+        else:
+            self.target_pressure = None
+        self.barostat_time_constant = barostat_time_constant
 
     def __str__(self):
         return f"Trajectory({len(self.frames)} frames)"
@@ -192,11 +231,68 @@ class Trajectory():
 
         return self
 
-    def initialize(self):
+    def initialize(self, frame=None, positions=None, velocities=None, accelerations=None, init_atoms=None, **kwargs):
         """
-        Adds first frame with randomly-initialized velocities. Should call ``self.save()``.
+        Adds first frame with randomly-initialized velocities.
+        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.
+
+        Can pass either a frame (from a different trajectory) or positions/velocities/accelerations.
+
+        Args:
+            frame (presto.frame.Frame): frame to modify or copy
+            positions (cctk.OneIndexedArray): starting positions
+            velocities (cctk.OneIndexedArray): starting velocities, optional.
+            accelerations (cctk.OneIndexedArray): starting accelerations, optional.
+            init_atoms (list of int): atoms to randomly give starting velocity. can be a list of indices.
+                if velocity is ``None`` all active atoms will be given a starting velocity.
+
+        Returns:
+            frame
         """
-        pass
+
+        # have we already initialized things?
+        if len(self.frames):
+            return
+        elif self.has_checkpoint():
+            self.load_from_checkpoint(slice(-1,None,None))
+            assert len(self.frames), "didn't load frames properly!"
+            return
+
+        logger.info("Initializing new trajectory...")
+
+        # if we get given a frame, we'll just copy everything from that
+        if frame is not None:
+            assert isinstance(frame, presto.frame.Frame), "need a valid frame"
+            positions = frame.positions
+            velocities = frame.velocities
+            accelerations = frame.accelerations
+
+        # initialize with zero velocity and acceleration
+        assert isinstance(positions, cctk.OneIndexedArray), "positions must be a one-indexed array!"
+        zeros = np.zeros_like(positions, dtype="float").view(cctk.OneIndexedArray)
+        # horrible bugfix - need to not pass the same object to velocities and accelerations!!
+        # ccw 7.6.22
+        frame = presto.frame.Frame(self, positions, copy.deepcopy(zeros), copy.deepcopy(zeros), bath_temperature=self.bath_scheduler(0), time=0.0)
+
+        # then adjust velocity and acceleration after-the-fact
+        if velocities is None:
+            frame.add_thermal_energy()
+        else:
+            assert isinstance(velocities, cctk.OneIndexedArray)
+            velocities[frame.inactive_mask()] = 0.0
+            frame.velocities += velocities
+
+            # add extra kick to requested atoms
+            if init_atoms:
+                frame.add_thermal_energy(atoms=init_atoms)
+
+        if accelerations is not None:
+            assert isinstance(accelerations, cctk.OneIndexedArray)
+            accelerations[frame.inactive_mask()] = 0.0
+            frame.accelerations += accelerations
+
+        self.frames = [frame]
+        self.save()
 
     def has_checkpoint(self):
         if self.checkpoint_filename is None:
@@ -206,12 +302,13 @@ class Trajectory():
         else:
             return False
 
-    def load_from_checkpoint(self, frames=slice(None)):
+    def load_from_checkpoint(self, frames="all"):
         """
         Loads frames from ``self.checkpoint_filename``.
 
         Args:
             frames (Slice object): if not all frames are desired, a Slice object can be passed
+                or a string - ``all``, ``first``, ``last``, or ``buffer``
 
         Returns:
             nothing
@@ -219,6 +316,17 @@ class Trajectory():
         if not self.has_checkpoint():
             return # nothing to load!
 
+        if frames == "all":
+            frames = slice(None)
+        elif frames == "first":
+            frames = slice(1, None, None)
+        elif frames == "last":
+            frames = slice(-1, None, None)
+        elif frames == "buffer":
+            frames = slice(-self.buffer, None, None)
+        else:
+            assert isinstance(frames, slice), "load_frames must be ``all``, ``first``, ``last``, or slice"
+
         self.initialize_lock()
         self.lock.acquire()
 
@@ -241,14 +349,21 @@ class Trajectory():
                 temperatures = h5.get("bath_temperatures")[frames]
 
                 # v0.2.2 - provisionally removing this
-#                all_times = None
-#                try:
-                all_times = h5.get("all_times")[frames]
-#                except Exception as e:
-#                    all_times = np.arange(0, self.timestep*len(all_energies)*self.save_interval, self.timestep*self.save_interval)
-#                    # this was added recently, so may be some backwards compatibility issues.
+                all_times = None
+                try:
+                    all_times = h5.get("all_times")[frames]
+                except Exception as e:
+                    all_times = np.arange(0, self.timestep*len(all_energies)*self.save_interval, self.timestep*self.save_interval)
+                    # this was added recently, so may be some backwards compatibility issues.
 #                    pass
 
+                # v0.2.6 - adding barostat support, working on backwards compatibility
+                all_scales = None
+                try:
+                    all_scales = h5.get("scale_factors")[frames]
+                except Exception as e:
+                    all_scales = np.ones_like(all_energies, dtype="float")
+
                 if isinstance(all_energies, np.ndarray):
                     assert len(all_positions) == len(all_energies)
                     assert len(all_velocities) == len(all_energies)
@@ -264,6 +379,7 @@ class Trajectory():
                         energy=all_energies[i],
                         bath_temperature=temperatures[i],
                         time=all_times[i],
+                        scale_factor=all_scales[i],
                     ))
 
         logger.info(f"Loaded trajectory from checkpoint file {self.checkpoint_filename} -- {len(self.frames)} frames read.")
@@ -291,14 +407,13 @@ class Trajectory():
             with h5py.File(self.checkpoint_filename, "r+") as h5:
                 n_atoms = len(self.atomic_numbers)
                 h5.attrs['finished'] = self.finished
+                h5.attrs['forwards'] = self.forwards
+
+                #### change 3.22.22 - this way masses save automatically, so you can edit isotopologues.
+                h5.attrs['masses'] = self.masses.view(np.ndarray)
 
                 all_energies = h5.get("all_energies")
                 old_n_frames = len(all_energies)
-                # v0.2.2 - provisionally removing this
-#                if "all_times" not in h5:
-#                    # time is a new column, so old checkpoints may not have it.
-#                    old_times = np.arange(0, self.timestep, self.timestep*old_n_frames)
-#                    h5.create_dataset("all_times", data=old_times, maxshape=(None,), compression="gzip", compression_opts=9)
 
                 all_times = h5.get("all_times")
                 last_saved_time = all_times[-1]
@@ -348,6 +463,22 @@ class Trajectory():
                 all_temps.resize((now_n_frames,))
                 all_temps[-new_n_frames:] = new_temps
 
+                # v0.2.6 - adding scale factor for barostat
+                all_scales = None
+                try:
+                    all_scales = h5.get("scale_factors")
+                    assert all_scales is not None
+                except Exception as e:
+                    # this is a new column, so old checkpoints may not have it.
+                    old_scales = np.ones(shape=old_n_frames, dtype="float")
+                    h5.create_dataset("scale_factors", data=old_scales, maxshape=(None,), compression="gzip", compression_opts=9)
+                    all_scales = h5.get("scale_factors")
+
+                new_scales = np.stack([frame.scale_factor for frame in frames_to_add])
+                all_scales.resize((now_n_frames,))
+                all_scales[-new_n_frames:] = new_scales
+
+
             logger.info(f"Saving to existing checkpoint file {self.checkpoint_filename} ({new_n_frames} frames added; {last_run_time:.1f}/{self.stop_time:.1f} fs run in total)")
         else:
             with h5py.File(self.checkpoint_filename, "w") as h5:
@@ -381,14 +512,17 @@ class Trajectory():
                 temps = np.asarray([frame.bath_temperature for frame in frames_to_add])
                 h5.create_dataset("bath_temperatures", data=temps, maxshape=(None,), compression="gzip", compression_opts=9)
 
+                scales = np.asarray([frame.scale_factor for frame in frames_to_add])
+                h5.create_dataset("scale_factors", data=scales, maxshape=(None,), compression="gzip", compression_opts=9)
+
             logger.info(f"Saving to new checkpoint file {self.checkpoint_filename} ({len(frames_to_add)} frames added; {last_run_time:.1f}/{self.stop_time:.1f} fs run in total)")
         self.lock.release()
 
-        # lower memory usage
+        # lower memory usage by not keeping every frame in memory.
         if keep_all:
             pass
         else:
-            self.frames = [self.frames[-1]]
+            self.frames = self.frames[-self.buffer:]
 
     def write_movie(self, filename, solvents="all", idxs=None):
         """
@@ -432,32 +566,12 @@ class Trajectory():
         else:
             raise ValueError(f"error writing {filename}: this filetype isn't currently supported!")
 
-    def as_ensemble(self, idxs=None):
+    def as_ensemble(self, idxs):
         ensemble = cctk.ConformationalEnsemble()
         for frame in self.frames[:-1]:
             ensemble.add_molecule(frame.molecule(idxs), {"bath_temperature": frame.bath_temperature, "energy": frame.energy})
         return ensemble
 
-    @classmethod
-    def new_from_checkpoint(cls, checkpoint, frame):
-        """
-        Creates new trajectory from the given checkpoint file.
-
-        Args:
-            checkpoint (str): path to checkpoint file
-            frame (int): the index of the desired frame
-
-        Returns:
-            new ``Trajectory`` object
-        """
-        assert isinstance(frame, int), "need an integer frame number"
-
-        new_traj = cls(checkpoint_filename=checkpoint)
-        new_traj.load_from_checkpoint(idxs=frame)
-
-        assert len(new_traj.frames) == 1, "got too many frames!"
-        return new_traj
-
     def initialize_lock(self):
         """
         Create hidden lockfile to accompany ``.chk`` file.
@@ -477,196 +591,39 @@ class Trajectory():
         """ Get last finished time """
         return self.frames[-1].time
 
-class ReactionTrajectory(Trajectory):
+def join(traj1, traj2, check=True):
     """
-    Attributes:
-        termination_function (function): detects if first or last Frame has reached product/SM or should otherwise be halted.
-            takes ``Frame`` argument as option and returns ``True``/``False``.
-            optionally, can return 1 for forward and 2 for reverse (to differentiate recrossing from productive).
-        time_after_finished (float/int): how long (in fs) to continue propagation after termination conditions reached
-    """
-
-    @classmethod
-    def new_from_checkpoint(self):
-        pass
-
-    def __init__(self, termination_function=None, time_after_finished=100, **kwargs):
-        super().__init__(**kwargs)
-
-        assert isinstance(time_after_finished, (int, float)), "time_after_finished must be numeric"
-        self.time_after_finished = time_after_finished
-        self.elapsed_since_finished = 0
-
-        assert hasattr(termination_function, "__call__"), "termination_function must be a function!"
-        self.termination_function = termination_function
-
-    def __str__(self):
-        return f"ReactionTrajectory({len(self.frames)} frames)"
-
-    def __repr__(self):
-        return f"ReactionTrajectory({len(self.frames)} frames)"
-
-    def initialize(self, frame=None, positions=None, velocities=None, accelerations=None, bath_temp=None, new_velocities=None, **kwargs):
-        """
-        Generates initial frame object for reaction trajectory. Initializes any non-zero velocities.
-        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.
-
-        Can pass either a frame object, or the relevant attributes (frame gets precedence).
-
-        Args:
-            frame (presto.frame.Frame): equilibrated frame
-            new_velocities (cctk.OneIndexedArray): array of velocities to add to equilibrated frame (for previously-frozen atoms)
-            velocities, accelerations, positions (cctk.OneIndexedArray): values from frame (instead of frame object)
-            bath_temperature (float):
-
-        Returns:
-            frame
-        """
-        logger.info("Initializing new reaction trajectory...")
-        if self.has_checkpoint():
-            self.load_from_checkpoint(slice(-1,None,None))
-            return
-
-        if frame is not None:
-            assert isinstance(frame, presto.frame.Frame), "need a valid frame"
-
-            positions = frame.positions
-            velocities = frame.velocities
-            accelerations = frame.accelerations
-            bath_temp = frame.bath_temperature
-
-        else:
-            assert positions is not None, "no Frame supplied, need positions"
-            assert velocities is not None, "no Frame supplied, need velocities"
-            assert accelerations is not None, "no Frame supplied, need accelerations"
-            assert bath_temp is not None, "no Frame supplied, need bath temperature"
-
-            assert isinstance(positions, cctk.OneIndexedArray)
-            assert isinstance(velocities, cctk.OneIndexedArray)
-            assert isinstance(accelerations, cctk.OneIndexedArray)
-            assert isinstance(bath_temp, (float, int, np.integer))
-
-        new_frame = presto.frame.Frame(self, positions, velocities, accelerations, bath_temperature=bath_temp, time=0.0)
-
-        if new_velocities is None:
-            random_gaussian = np.random.normal(size=positions.shape).view(cctk.OneIndexedArray)
-            random_gaussian[new_frame.active_mask()] = 0
-            new_velocities = random_gaussian * np.sqrt(bath_temp * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))
-
-        new_frame.velocities += new_velocities.view(cctk.OneIndexedArray)
-
-        self.frames = [new_frame]
-        self.save()
-
-class EquilibrationTrajectory(Trajectory):
-    """
-    Attributes:
-        bath_scheduler (function): takes current time and returns target temperature
-            alternatively, pass a number if you want a constant-temperature bath
-    """
-
-    def __init__(self, **kwargs):
-        super().__init__(**kwargs)
-
-        bath_scheduler = kwargs.get("bath_scheduler")
-
-        if hasattr(bath_scheduler, "__call__"):
-            self.bath_scheduler = bath_scheduler
-        elif isinstance(bath_scheduler, (int, float)):
-            def sched(time):
-                return bath_scheduler
-
-            self.bath_scheduler = sched
-        else:
-            raise ValueError(f"unknown type {type(bath_scheduler)} for bath_scheduler - want either a function or a number!")
-
-    def __str__(self):
-        return f"EquilibrationTrajectory({len(self.frames)} frames)"
-
-    def __repr__(self):
-        return f"EquilibrationTrajectory({len(self.frames)} frames)"
-
-    def initialize(self, positions, velocities=None, accelerations=None, **kwargs):
-        """
-        Generates initial frame object for initialization trajectory.
-        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.
-
-        Args:
-            positions (cctk.OneIndexedArray): starting positions
-            velocities (cctk.OneIndexedArray): starting velocities, optional.
-            accelerations (cctk.OneIndexedArray): starting accelerations, optional.
-
-        Returns:
-            frame
-        """
-        logger.info("Initializing new equilibration trajectory...")
-        if self.has_checkpoint():
-            self.load_from_checkpoint(slice(-1,None,None))
-            assert len(self.frames) == 1, "didn't load frames properly!"
-            return
-
-        assert isinstance(positions, cctk.OneIndexedArray), "positions must be a one-indexed array!"
-
-        # determine active atoms
-        inactive_mask = np.ones(shape=len(positions)).view(cctk.OneIndexedArray)
-        inactive_mask[self.active_atoms] = 0
-        inactive_mask = inactive_mask.astype(bool)
-
-        if velocities is None:
-            # add random velocity to everything
-            sigma = np.sqrt(self.bath_scheduler(0) * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))
-            velocities = np.random.normal(scale=sigma, size=positions.shape).view(cctk.OneIndexedArray)
-            velocities[inactive_mask] = 0
-
-            # subtract out center-of-mass translational motion
-            com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
-            correction_tran = np.tile(com_translation / np.sum(self.masses[self.active_atoms]), (len(velocities),1))
-            correction_tran[inactive_mask] = 0
-            velocities = velocities - correction_tran
-            com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
-            assert np.linalg.norm(np.sum(self.masses.reshape(-1,1) * velocities, axis=0)) < 0.0001, "didn't remove COM translation well enough!"
-
-            velocities = velocities.view(cctk.OneIndexedArray)
-        else:
-            assert isinstance(velocities, cctk.OneIndexedArray)
-            velocities[inactive_mask] = 0
-
-        if accelerations is None:
-            accelerations = np.zeros_like(positions).view(cctk.OneIndexedArray)
-        else:
-            assert isinstance(accelerations, cctk.OneIndexedArray)
-            accelerations[inactive_mask] = 0
-
-        self.frames = [presto.frame.Frame(self, positions, velocities, accelerations, bath_temperature=self.bath_scheduler(0), time=0.0)]
-        self.save()
-
-def join(traj1, traj2):
-    """
-    Join two reaction trajectories together -- one forward and one reverse!
+    Join two trajectories together -- one forward and one reverse!
     Returns a single reaction trajectory.
-
+    Eugene edit: Times for the reverse trajectory will be made negative.
     Args:
-        traj1 (presto.ReactionTrajectory): forward trajectory
-        traj2 (presto.ReactionTrajectory): reverse trajectory
-
+        traj1 (presto.Trajectory): forward trajectory
+        traj2 (presto.Trajectory): reverse trajectory
+        check (bool): check if the joining makes sense (highly recommended)
     Returns:
         combined ``ReactionTrajectory``
     """
     logger.info("Joining forward and reverse reaction trajectories....")
-    assert isinstance(traj1, ReactionTrajectory), "Need a ReactionTrajectory"
-    assert isinstance(traj2, ReactionTrajectory), "Need a ReactionTrajectory"
+    assert isinstance(traj1, Trajectory), "Need a ReactionTrajectory"
+    assert isinstance(traj2, Trajectory), "Need a ReactionTrajectory"
 
     assert traj1.forwards == True, "First trajectory must be forwards!"
     assert traj2.forwards == False, "Second trajectory must be reverse!"
 
-    assert traj1.finished, "First trajectory must be finished!"
-    assert traj2.finished, "Second trajectory must be finished!"
+    # Eugene edit: there doesn't seem to be a need to have the trajectories be finished
+    #assert traj1.finished, "First trajectory must be finished!"
+    #assert traj2.finished, "Second trajectory must be finished!"
 
-    assert np.array_equal(traj1.frames[0].positions, traj2.frames[0].positions), "Link positions must be same!"
-    assert np.array_equal(traj1.frames[0].velocities, traj2.frames[0].velocities), "Link velocities must be same!"
-    assert np.array_equal(traj1.frames[0].accelerations, traj2.frames[0].accelerations), "Link accelerations must be same!"
+    # load all frames
+    traj1.load_from_checkpoint()
+    traj2.load_from_checkpoint()
 
-    new_traj = ReactionTrajectory(
+    if check:
+        assert np.array_equal(traj1.frames[0].positions, traj2.frames[0].positions), "Link positions must be same!"
+        assert np.array_equal(traj1.frames[0].velocities, traj2.frames[0].velocities), "Link velocities must be same!"
+        assert np.array_equal(traj1.frames[0].accelerations, traj2.frames[0].accelerations), "Link accelerations must be same!"
+
+    new_traj = Trajectory(
         timestep = traj1.timestep,
         atomic_numbers = traj1.atomic_numbers,
         high_atoms = traj1.high_atoms,
@@ -682,11 +639,18 @@ def join(traj1, traj2):
     r_frames = copy.deepcopy(traj2.frames)
     r_frames.reverse()
 
+    # Eugene edit: make reverse reverse frame times negative
+    for frame in r_frames:
+        frame.time = -frame.time
+
     new_traj.frames = r_frames + f_frames[1:] # don't double-count middle frame
 
-    if traj1.finished == 2:
+    if traj1.finished == 2 and traj2.finished == 1:
         #### if the first traj finished with the reverse condition, reverse order
         new_traj.frames = new_traj.frames[::-1]
 
-    return new_traj
+        # Eugene edit: if this happens, reverse all times
+        for frame in new_traj.frames:
+            frame.time = -frame.time
 
+    return new_traj
